/*
 * HttpD.c
 *
 *      Author: Trax
 */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <avr/pgmspace.h>
#include "HttpD.h"
#include "../uip/uip.h"
#include "../../Ports.h"
#include "../../Options.h"
#include "../../Screens.h"
#include "../../shell.h"

extern char g_Buffer[128];

int ParseHeader(char* pBuff, int iLen, tTrans* Trans);
int MakeHttpData(char* pBuff, int nMSS, tTrans* Trans);

#define STATE_CONNECTED		0
#define STATE_SENDHEADER	1
#define STATE_HEADERSENT	2
#define STATE_SENDDATA		3
#define STATE_DATASENT		4

uint8_t strcopy_P(char** pBuff, int* nMSS, int* iPos, const prog_char* pSrc);

static const prog_char g_HttpHeader[] =
				"HTTP/1.1 200 OK\r\n"
				"Content-Type:text/html\r\n"
				"Connection:close\r\n"
				"Transfer-Encoding:chunked\r\n"
				"\r\n";

void HttpDInit()
{
	//Start listening on our port
	uip_listen(HTONS(PORT_HTTPD));
}

void SendHttpData(char* DATA, int MSS, tTrans* Trans)
{
	uint16_t nBufSize = MakeHttpData(DATA+5, MSS-7, Trans);
	if (nBufSize == 0)
	{
		Trans->State = STATE_DATASENT;
		memcpy_P(uip_appdata,PSTR("0\r\n\r\n"),5);
		nBufSize = 5;
	}
	else
	{
		memcpy_P(uip_appdata,PSTR("   \r\n"),5);
		memcpy(uip_appdata,Trans->Buff,sprintf_P(Trans->Buff,PSTR("%x"),nBufSize));
		memcpy_P(DATA+5+nBufSize,PSTR("\r\n"),2);
		nBufSize += 7;
	}
	uip_send(uip_appdata, nBufSize);
}

void HttpDCall(char* pBuffer, int nBytes, tTrans* pSocket)
{
	if (uip_connected())
	{
		pSocket->State = STATE_CONNECTED;
		pSocket->Index = 0;
		pSocket->h=pSocket->i=pSocket->j=pSocket->k=pSocket->l=pSocket->m=0;
	}
	else if (uip_newdata() || uip_acked())
	{
		if (pSocket->State == STATE_CONNECTED)
		{
			if(ParseHeader(pBuffer,nBytes,pSocket) == 0)
				pSocket->State = STATE_SENDHEADER;
		}

		if (pSocket->State == STATE_SENDHEADER)
		{
			pSocket->State = STATE_HEADERSENT;
			strcpy_P(uip_appdata, g_HttpHeader);
			uip_send(uip_appdata,strlen_P(g_HttpHeader));
			return;
		}

		if(pSocket->State == STATE_HEADERSENT)
		{
			pSocket->State = STATE_SENDDATA;
		}

		if (pSocket->State == STATE_SENDDATA)
		{
			// backup state
			pSocket->h_ = pSocket->h;
			pSocket->i_ = pSocket->i;
			pSocket->j_ = pSocket->j;
			pSocket->k_ = pSocket->k;
			pSocket->l_ = pSocket->l;
			pSocket->m_ = pSocket->m;

			SendHttpData(uip_appdata, uip_mss(), pSocket);
		}
		else if(pSocket->State == STATE_DATASENT)
			uip_close();
	}
	else if (uip_rexmit())
	{
		if (pSocket->State == STATE_HEADERSENT)
		{
			strcpy_P(uip_appdata, g_HttpHeader);
			uip_send(uip_appdata,strlen_P(g_HttpHeader));
		}
		else if (pSocket->State == STATE_SENDDATA)
		{
			// restore state
			pSocket->h = pSocket->h_;
			pSocket->i = pSocket->i_;
			pSocket->j = pSocket->j_;
			pSocket->k = pSocket->k_;
			pSocket->l = pSocket->l_;
			pSocket->m = pSocket->m_;

			SendHttpData(uip_appdata, uip_mss(), pSocket);
		}
		else if(pSocket->State == STATE_DATASENT)
		{
			memcpy_P(uip_appdata,PSTR("0\r\n\r\n"),5);
			uip_send(uip_appdata, 5);
		}
	}
}



// Combo box Formating functions

uint8_t strcopy_C(char** pBuff, int* nMSS, int* iPos, const char* pSrc)
{
	char Char;
	while((Char = pSrc[*iPos]) != '\0')
	{
		**pBuff = Char;
		(*pBuff)++;
		(*iPos)++;
		if(--*nMSS == 0)
			return 1;
	}
	*iPos = 0;
	return 0;
}

uint8_t strcopy_P(char** pBuff, int* nMSS, int* iPos, const prog_char* pSrc)
{
	char Char;
	while ((Char = pgm_read_byte(pSrc+*iPos)) != '\0')
	{
		**pBuff = Char;
		(*pBuff)++;
		(*iPos)++;
		if(--*nMSS == 0)
			return 1;
	}
	*iPos = 0;
	return 0;
}

uint8_t strprint(char** pBuff, int* nMSS, int* iPos, const prog_char * _Format, ...)
{
	va_list argptr;
	va_start(argptr, _Format);
	vsnprintf_P(g_Buffer, ARRSIZE(g_Buffer), _Format, argptr);
	va_end(argptr);

	return strcopy_C(pBuff, nMSS, iPos, g_Buffer);
}

__inline int OP_Chk(const prog_char* Str, char op)
{
	return (pgm_read_byte(Str) == '%' && pgm_read_byte(Str+1) == op);
}

// Common HTML Header:
static const prog_char g_HtmlHeaderBegin[] =
				"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\r\n"
				"<html><head>\r\n"
				"<title>AVR Smart WebSrv by Trax Xavier</title>\r\n";


static const prog_char g_HtmlHeaderStyle[] =
				"<html><head>\r\n"
				"<title>AVR Smart WebSrv by Trax Xavier</title>\r\n"
				"<style type='text/css'>\r\n"
				".s1 { background-color: #FF7777; }\r\n"
				".s2 { background-color: #77FF77; }\r\n"
				".s3 { background-color: #7777FF; }\r\n"
				".s4 { background-color: #FFFF77; }\r\n"
				".s5 { background-color: #FFFFFF; }\r\n"
				"td { text-align: center; }\r\n"
				"</style>\r\n";

static const prog_char g_HtmlHeaderScript[] =
				"<script type='text/javascript'>\r\n"
				"function genCombo(list , name, sel , submit) {\r\n"
				"   var s = '<select name=\\'' + name + '\\'';\r\n"
				"   if(submit != '')\r\n"
				"      s += 'onchange=\\'' + submit + '.submit();\\'';\r\n"
				"   s += '>';\r\n"
				"   for (var i = 0; i < list.length; i++) {\r\n"
				"      s += '<option value=\\'' + list[i] + '\\'';\r\n"
				"      if (list[i] == sel)\r\n"
				"         s += ' selected=\\'selected\\'';\r\n"
				"      s += '>' + list[i] + '</option>';\r\n"
				"   }\r\n"
				"   document.write(s + '</select>');\r\n"
				"}\r\n"
				"</script>\r\n";

static const prog_char g_HtmlHeaderEnd[] =
				"</head><body>\r\n";

static const prog_char g_HtmlFooter[] =
				"</body></html>\r\n";


// Combo box FOrmating functions
static const prog_char g_IOPortsA[] = "IOPortsA";
static const prog_char g_IOPortsP[] = "IOPortsP";
static const prog_char g_IOPortsM[] = "IOPortsM";
static const prog_char g_IOPorts[] = "IOPorts";
static const prog_char g_IOPorts2[] = "IOPorts2";
static const prog_char g_APorts[] = "APorts";
static const prog_char g_OPorts[] = "OPorts";
static const prog_char g_BinV[] = "BinV";
static const prog_char g_BitV[] = "BitV";
static const prog_char g_AdcV[] = "AdcV";
static const prog_char g_LCDUse[] = "LCDSupport";
static const prog_char g_LCDPorts[] = "LCDPorts";

static const prog_char g_HtmlOptsBegin[] =
				"<script type='text/javascript'> var %S = new Array('None'";

static const prog_char g_HtmlOpts[] =
				", '%s'";

static const prog_char g_HtmlOptsEnd[] =
				");</script>\r\n";

static const prog_char g_HtmlCBNo[] =
				"<noscript><input name='%S%d' value='%s' maxlength='8' size='2' type='text'></noscript>";

static const prog_char g_HtmlCBScript[] =
				"<script type='text/javascript'>genCombo(%S, '%S%d', '%s', '%S');</script>";

int PrintPortsList(char** pBuff, int* nMSS, int* iStage, int* iStep, int* iPos, const prog_char* pName, int Type)
{
	if(*iStage == 0)
	{
		if(strprint(pBuff, nMSS, iPos, g_HtmlOptsBegin, pName))
			return 1;
		*iStage = 1;
	}
	if(*iStage == 1)
	{
		const int PortCount = GetPortCount();
		for(;(*iStep)<PortCount;(*iStep)++)
		{
			int IO = GetPortIO(*iStep);
			if(Type == 1 && IO != OUTPUT && IO != INPUT && IO != INPUT_ADC && IO != OUTPUT_PWM && IO != MEMORY && IO != MEMORY_SW)
				continue;
			else if(Type == 2 && IO != OUTPUT && IO != INPUT && IO != MEMORY && IO != MEMORY_SW && IO)
				continue;
			else if(Type == 3 && IO != OUTPUT && IO != OUTPUT_PWM && IO != MEMORY && IO != MEMORY_SW)
				continue;
			else if (Type == 4 && IO != INPUT_ADC && IO != OUTPUT_PWM)
				continue;
			else if (Type == 5 && IO != OUTPUT && IO != MEMORY && IO != MEMORY_SW)
				continue;

			char* ID = GetPortID(*iStep);
			if(strprint(pBuff, nMSS, iPos, g_HtmlOpts, ID))
				return 1;
		}
		*iStep = 0;
		*iStage = 2;
	}
	if(*iStage == 2)
	{
		if(strcopy_P(pBuff, nMSS, iPos, g_HtmlOptsEnd))
			return 1;
		*iStage = 0;
	}
	return 0;
}

int PrintOptionList(char** pBuff, int* nMSS, int* iStage, int* iStep, int* iPos, const prog_char* pName, const prog_char* pOptions)
{
	if(*iStage == 0)
	{
		if(strprint(pBuff, nMSS, iPos, g_HtmlOptsBegin, pName))
			return 1;
		*iStage = 1;
	}
	if(*iStage == 1)
	{
		int i;
		for(i=0; pgm_read_byte(pOptions) != 0; i++)
		{
			if(i == *iStep)
			{
				if(strcmp_P(".",pOptions) != 0)
				{
					char Option[16];
					strcpy_P(Option,pOptions);
					if(strprint(pBuff, nMSS, iPos, g_HtmlOpts, Option))
						return 1;
				}
				(*iStep)++;
			}
			pOptions += strlen_P(pOptions)+1;
		}
		*iStep = 0;
		*iStage = 2;
	}
	if(*iStage == 2)
	{
		if(strcopy_P(pBuff, nMSS, iPos, g_HtmlOptsEnd))
			return 1;
		*iStage = 0;
	}
	return 0;
}

static const char g_None[] = "None";
const char* GetComboOpt(int pOpt, const prog_char* pOptions, char* pBuff)
{
	int Sel = 0;
	while(pgm_read_byte(pOptions) != 0)
	{
		if(Sel == pOpt)
		{
			strcpy_P(pBuff,pOptions);
			return pBuff;
		}
		Sel++;
		pOptions += strlen_P(pOptions)+1;
	}
	return g_None;
}

int GetComboVal(const char* pTest, const prog_char* pOptions)
{
	int Sel = 0;
	while(pgm_read_byte(pOptions) != 0)
	{
		if(strcmp_P(pTest,pOptions) == 0)
			return Sel;
		pOptions += strlen_P(pOptions)+1;
		Sel++;
	}
	return -1;
}

extern const prog_char g_NULL[];
int PrintCombo(char** pBuff, int* nMSS, int* iStage, int* iPos, const prog_char* pName, const prog_char* pID, int Index, const char* Sel, const prog_char* Submit)
{
	if(Sel == NULL)
		Sel = g_None;
	if(*iStage == 0)
	{
		if(strprint(pBuff, nMSS, iPos, g_HtmlCBNo, pID, Index, Sel))
			return 1;
		*iStage = 1;
	}
	if(*iStage == 1)
	{
		if(strprint(pBuff, nMSS, iPos, g_HtmlCBScript, pName, pID, Index, Sel, Submit))
			return 1;
		*iStage = 0;
	}
	return 0;
}


// HW Config
static const prog_char g_CfgHtmlForm[] =
				"<h1>Hardware Setup</h1>\r\n<form name='frmCfg'  method='POST' action='/Cfg'>\r\n";

static const prog_char g_CfgHtmlTable1[] =
				"<table border='0' cellpadding='0' cellspacing='0'>"
				"<tr><td valign='top'><table border='1px'><colgroup><col width='40px'><col width='140px'><col width='40px'></colgroup>\r\n"
				"<td>ID</td><td>Port Label</td><td>I/O</td>\r\n";

static const prog_char g_CfgHtmlTabel2[] =
				"</table></td><td >&nbsp;</td><td valign='top'><table border='1px'><colgroup><col width='40px'><col width='140px'><col width='40px'></colgroup>\r\n"
				"<td>ID</td><td>Memory Label</td><td>State</td>\r\n";

static const prog_char g_CfgHtmlLine[] =
				"<tr><td class='s%C'>%P</td><td class='s%C'><input name='Label%N' value='%L' maxlength='20' size='20' type='text'></td><td class='s%C'>%X</td></tr>\r\n";
const prog_char g_OptsIOA[] =
				"Output\0Input\0ADC\0.\0";
const prog_char g_OptsIO[] =
				"Output\0Input\0";
const prog_char g_OptsIOP[] =
				"Output\0Input\0.\0PWM\0";
const prog_char g_OptsMEM[] =
				".\0.\0.\0.\0Hidden\0Visible\0";
const prog_char g_OptsALL[] =
				"Output\0Input\0ADC\0PWM\0Hidden\0Visible\0";

static const prog_char g_HtmlEmpty[] =
				"&nbsp;";

static const prog_char g_CfgHtmlTabelEnd[] =
				"</td></tr></table></table><br><br>LCD Support:&nbsp;";

const prog_char g_OptsLCD[] =
				"Default\0Custom\0";

static const prog_char g_CfgHtmlLines[] =
				"&nbsp;&nbsp;&nbsp;Lines: <input name='Lines' value='%d' maxlength='1' size='2' type='text'>";

static const prog_char g_CfgHtmlLCD[] =
				"<br><br><table border='1px'><tr><td class='s1'>RS</td><td class='s3'>EN</td><td class='s2'>D4</td><td class='s2'>D5</td><td class='s2'>D6</td><td class='s2'>D7</td><td class='s4'>SL</td><td class='s4'>SR</td></tr>\r\n"
				"<tr><td class='s1'>%2</td><td class='s3'>%3</td><td class='s2'>%4</td><td class='s2'>%5</td><td class='s2'>%6</td><td class='s2'>%7</td><td class='s4'>%8</td><td class='s4'>%9</td></tr></table>\r\n";

static const prog_char g_CfgHtmlFormEnd[] =
				"\r\n<br><br><input name='Update' type='submit' value='Update'>&nbsp;<input name='Save0' type='submit' value='Save'>\r\n"
				"</form>\r\n";

static const prog_char g_frmCfg[] = "frmCfg";

extern const prog_char g_Support[];
extern const prog_char g_IO[];
extern const prog_char g_LCDPin[];

int CfgHttpData(char* pBuff, int nMSS, tTrans* Trans)
{
	char* pBuff0 = pBuff;

	if(Trans->h == 0)
	{
		if(Trans->j == 0)
		{
			if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_HtmlHeaderBegin))
				return pBuff-pBuff0;
			Trans->j = 1;
		}

		if(Trans->j == 1)
		{
			if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_HtmlHeaderStyle))
				return pBuff-pBuff0;
			Trans->j = 2;
		}

		if(Trans->j == 2)
		{
			if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_HtmlHeaderScript))
				return pBuff-pBuff0;
			Trans->j = 3;
		}

		if(Trans->j == 3)
		{
			if(PrintOptionList(&pBuff, &nMSS, &Trans->k, &Trans->l, &Trans->m, g_IOPorts, g_OptsIO))
				return pBuff-pBuff0;
			Trans->j = 4;
		}

		if(Trans->j == 4)
		{
			if(PrintOptionList(&pBuff, &nMSS, &Trans->k, &Trans->l, &Trans->m, g_IOPortsA, g_OptsIOA))
				return pBuff-pBuff0;
			Trans->j = 5;
		}

		if(Trans->j == 5)
		{
			if(PrintOptionList(&pBuff, &nMSS, &Trans->k, &Trans->l, &Trans->m, g_IOPortsP, g_OptsIOP))
				return pBuff-pBuff0;
			Trans->j = 6;
		}

		if(Trans->j == 6)
		{
			if(PrintOptionList(&pBuff, &nMSS, &Trans->k, &Trans->l, &Trans->m, g_IOPortsM, g_OptsMEM))
				return pBuff-pBuff0;
			Trans->j = 7;
		}

		if(Trans->j == 7)
		{
			if(GetLCDSupport() == 2)
			{
				if(PrintPortsList(&pBuff, &nMSS, &Trans->k, &Trans->l, &Trans->m, g_LCDPorts, 0))
					return pBuff-pBuff0;
			}
			Trans->j = 8;
		}

		if(Trans->j == 8)
		{
			if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_HtmlHeaderEnd))
				return pBuff-pBuff0;
			Trans->j = 0;
		}

		Trans->h = 1;
	}

	if(Trans->h == 1)
	{
		if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_CfgHtmlForm))
			return pBuff-pBuff0;
		Trans->h = 2;
	}

	if(Trans->h == 2)
	{
		if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_CfgHtmlTable1))
			return pBuff-pBuff0;
		Trans->h = 3;
	}

	if(Trans->h == 3)
	{
		const int CtrlHtmlLineLen = strlen_P(g_CfgHtmlLine);
		const int PortCount = GetPortCount();
		const int HWPortCount= GetHWPortCount();

		while(Trans->i<PortCount)
		{
			if(Trans->i == HWPortCount && Trans->l == 0)
			{
				if(strcopy_P(&pBuff, &nMSS, &Trans->k, g_CfgHtmlTabel2))
					return pBuff-pBuff0;
				Trans->l = 1;
			}

			while(Trans->j<CtrlHtmlLineLen)
			{
				if(OP_Chk(g_CfgHtmlLine+Trans->j,'N'))
				{
					itoa(Trans->i, g_Buffer, 10);
					if(strcopy_C(&pBuff, &nMSS, &Trans->k, g_Buffer))
						return pBuff-pBuff0;
					Trans->j+=2;
				}
				else if(OP_Chk(g_CfgHtmlLine+Trans->j,'P'))
				{
					if(strcopy_C(&pBuff, &nMSS, &Trans->k, GetPortID(Trans->i)))
						return pBuff-pBuff0;
					Trans->j+=2;
				}
				else if(OP_Chk(g_CfgHtmlLine+Trans->j,'L'))
				{
					char* Label = GetPortLabel(Trans->i,0);
					if(Label)
					{
						if(strcopy_C(&pBuff, &nMSS, &Trans->k, Label))
							return pBuff-pBuff0;
					}
					Trans->j+=2;
				}
				else if(OP_Chk(g_CfgHtmlLine+Trans->j,'X'))
				{
					int ret = 0;
					uint8_t IO = GetPortIO(Trans->i);
					if(IO >= LC_RS && IO <= LC_SR)
						ret = strcopy_C(&pBuff, &nMSS, &Trans->k, GetLCDPin(IO-LC_RS));
					else if(GetPortADC(Trans->i) != NONE)
						ret = PrintCombo(&pBuff, &nMSS, &Trans->k, &Trans->m, g_IOPortsA, g_IO, Trans->i, GetComboOpt(IO,g_OptsIOA,Trans->Buff), g_NULL);
					else if(GetPortPWM(Trans->i) != NONE)
						ret = PrintCombo(&pBuff, &nMSS, &Trans->k, &Trans->m, g_IOPortsP, g_IO, Trans->i, GetComboOpt(IO,g_OptsIOP,Trans->Buff), g_NULL);
					else if(GetPortMEM(Trans->i) != NONE)
						ret = PrintCombo(&pBuff, &nMSS, &Trans->k, &Trans->m, g_IOPortsM, g_IO, Trans->i, GetComboOpt(IO,g_OptsMEM,Trans->Buff), g_NULL);
					else
						ret = PrintCombo(&pBuff, &nMSS, &Trans->k, &Trans->m, g_IOPorts, g_IO, Trans->i, GetComboOpt(IO,g_OptsIO,Trans->Buff), g_NULL);

					if(ret)
						return pBuff-pBuff0;
					Trans->j+=2;
				}
				else if(OP_Chk(g_CfgHtmlLine+Trans->j,'C'))
				{
					if(Trans->i < 8)		*pBuff++ = '1';
					else if(Trans->i < 16)	*pBuff++ = '2';
					else if(Trans->i < 24)	*pBuff++ = '3';
					else if(Trans->i < 25)	*pBuff++ = '4';
					else if(Trans->i < 33)	*pBuff++ = '5';
					Trans->j+=2;
					if(!--nMSS)
						return pBuff-pBuff0;
				}
				else
				{
					*pBuff++ = pgm_read_byte(g_CfgHtmlLine+Trans->j);
					Trans->j+=1;
					if(!--nMSS)
						return pBuff-pBuff0;
				}
			}
			Trans->j = 0;
			Trans->i++;
		}
		Trans->i = 0;
		Trans->l = 0;
		Trans->h = 4;
	}

	if(Trans->h == 4)
	{
		if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_CfgHtmlTabelEnd))
			return pBuff-pBuff0;
		Trans->h = 5;
	}

	if(Trans->h == 5)
	{
		if(Trans->j == 0)
		{
			if(PrintOptionList(&pBuff, &nMSS, &Trans->k, &Trans->l, &Trans->m, g_LCDUse, g_OptsLCD))
				return pBuff-pBuff0;
			Trans->j = 1;
		}

		if(Trans->j == 1)
		{
			if(PrintCombo(&pBuff, &nMSS, &Trans->k, &Trans->l, g_LCDUse, g_Support, 0, GetComboOpt(GetLCDSupport()-1,g_OptsLCD,Trans->Buff), g_frmCfg))
				return pBuff-pBuff0;
			Trans->j = 0;
		}
		Trans->h = 6;
	}

	if(Trans->h == 6)
	{
		if(GetLCDSupport())
		{
			if(strprint(&pBuff, &nMSS, &Trans->i, g_CfgHtmlLines, GetLCDLines()))
				return pBuff-pBuff0;
		}
		Trans->h = 7;
	}

	if(Trans->h == 7)
	{
		if(GetLCDSupport() == 2)
		{
			const int CfgHtmlLCDSetupLen = strlen_P(g_CfgHtmlLCD);

			while(Trans->i<CfgHtmlLCDSetupLen)
			{
				if(g_CfgHtmlLCD[Trans->i] == '%')
				{
					strncpy_P(g_Buffer,g_CfgHtmlLCD+Trans->i+1,1);
					int Pin = atoi(g_Buffer)-2;

					if(PrintCombo(&pBuff, &nMSS, &Trans->k, &Trans->m, g_LCDPorts, g_LCDPin, Pin, GetLCDPinID(Pin), g_frmCfg))
						return pBuff-pBuff0;

					Trans->i+=2;
				}
				else
				{
					*pBuff++ = pgm_read_byte(g_CfgHtmlLCD+Trans->i);
					Trans->i+=1;
					if(!--nMSS)
						return pBuff-pBuff0;
				}
			}
			Trans->i = 0;
		}
		Trans->h = 8;
	}

	if(Trans->h == 8)
	{
		if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_CfgHtmlFormEnd))
			return pBuff-pBuff0;
		Trans->h = 9;
	}


	if(Trans->h == 9)
	{
		if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_HtmlFooter))
			return pBuff-pBuff0;

		Trans->h = -1;
	}

	return pBuff-pBuff0;
}

// Option Config
static const prog_char g_MgmtHtmlForm[] =
				"<h1>Management</h1>\r\n<form name='frmMgmt'  method='POST' action='/Mgmt'>\r\n";

static const prog_char g_MgmtHtmlTable[] =
				"<table border='1'><colgroup><col width='40'><col width='40'><col width='40'><col width='40'><col width='40'><col width='110'></colgroup>\r\n"
				"<td>Input</td><td>Value</td><td>Optional</td><td>Output</td><td>Value</td><td>Position</td>\r\n";

static const prog_char g_MgmtHtmlLine[] =
				"<tr><td>%I</td><td>%J</td><td>%X</td><td>%O</td><td>%P</td><td><input value='U' type='submit' name='UpOP%N'><input value='R' type='submit' name='RemOP%N'><input value='D' type='submit' name='DownOP%N'></td></tr>\r\n";

const prog_char g_OptsBitV[] =
				"Lo\0Hi\0Lo.S\0Hi.S\0";
const prog_char g_OptsBinV[] =
				"Lo\0Hi\0AND\0NOT.AND\0AND.NOT\0NOR\0OR\0NOT.OR\0OR.NOT\0NAND\0XOR\0XNOR\0";
const prog_char g_OptsAdcV[] =
				"EQ.v\0GT.v\0LT.v\0NE.v\0EQ.c\0GT.c\0LT.c\0NE.c\0";


static const prog_char g_MgmtHtmlEB[] =
				"<input name='%s%d' onblur = 'frmMgmt.submit();' value='%d' maxlength='10' size='5' type='text'>";

static const prog_char g_MgmtHtmlFormEnd[] =
				"<tr><td>&nbsp;</td><td>&nbsp;</td><td>&nbsp;</td><td>&nbsp;</td><td>&nbsp;</td><td><input value='A' type='submit' name='AddOP'></td><tr></table>\r\n"
				"<br><input name='Update' type='submit' value='Update'>&nbsp;<input name='Save1' type='submit' value='Save'></form>\r\n";

static const prog_char g_frmMgmt[] = "frmMgmt";

extern const prog_char g_Input[];
extern const prog_char g_InVal[];
extern const prog_char g_InAux[];
extern const prog_char g_Output[];
extern const prog_char g_OutVal[];

int MgmtHttpData(char* pBuff, int nMSS, tTrans* Trans)
{
	char* pBuff0 = pBuff;

	if(Trans->h == 0)
	{
		if(Trans->j == 0)
		{
			if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_HtmlHeaderBegin))
				return pBuff-pBuff0;
			Trans->j = 1;
		}

		if(Trans->j == 1)
		{
			if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_HtmlHeaderScript))
				return pBuff-pBuff0;
			Trans->j = 2;
		}

		if(Trans->j == 2)
		{
			if(PrintPortsList(&pBuff, &nMSS, &Trans->k, &Trans->l, &Trans->m, g_IOPorts, 1))
				return pBuff-pBuff0;
			Trans->j = 3;
		}

		if(Trans->j == 3)
		{
			if(PrintPortsList(&pBuff, &nMSS, &Trans->k, &Trans->l, &Trans->m, g_IOPorts2, 2))
				return pBuff-pBuff0;
			Trans->j = 4;
		}

		if(Trans->j == 4)
		{
			if(PrintPortsList(&pBuff, &nMSS, &Trans->k, &Trans->l, &Trans->m, g_OPorts, 3))
				return pBuff-pBuff0;
			Trans->j = 5;
		}

		if(Trans->j == 5)
		{
			if(PrintPortsList(&pBuff, &nMSS, &Trans->k, &Trans->l, &Trans->m, g_APorts, 4))
				return pBuff-pBuff0;
			Trans->j = 6;
		}

		if(Trans->j == 6)
		{
			if(PrintOptionList(&pBuff, &nMSS, &Trans->k, &Trans->l, &Trans->m, g_BitV, g_OptsBitV))
				return pBuff-pBuff0;
			Trans->j = 7;
		}

		if(Trans->j == 7)
		{
			if(PrintOptionList(&pBuff, &nMSS, &Trans->k, &Trans->l, &Trans->m, g_BinV, g_OptsBinV))
				return pBuff-pBuff0;
			Trans->j = 8;
		}

		if(Trans->j == 8)
		{
			if(PrintOptionList(&pBuff, &nMSS, &Trans->k, &Trans->l, &Trans->m, g_AdcV, g_OptsAdcV))
				return pBuff-pBuff0;
			Trans->j = 9;
		}

		if(Trans->j == 9)
		{
			if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_HtmlHeaderEnd))
				return pBuff-pBuff0;
			Trans->j = 0;
		}
		Trans->h = 1;
	}

	if(Trans->h == 1)
	{
		if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_MgmtHtmlForm))
			return pBuff-pBuff0;
		Trans->h = 2;
	}

	if(Trans->h == 2)
	{
		if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_MgmtHtmlTable))
			return pBuff-pBuff0;
		Trans->h = 3;
	}

	if(Trans->h == 3)
	{
		const int MgmtHtmlLineLen = strlen_P(g_MgmtHtmlLine);
		const int OptCount = GetOptCount();

		while(Trans->i<OptCount)
		{
			while(Trans->j<MgmtHtmlLineLen)
			{
				if(OP_Chk(g_MgmtHtmlLine+Trans->j,'N'))
				{
					itoa(Trans->i, g_Buffer, 10);
					if(strcopy_C(&pBuff, &nMSS, &Trans->k, g_Buffer))
						return pBuff-pBuff0;
					Trans->j+=2;
				}
				else if(OP_Chk(g_MgmtHtmlLine+Trans->j,'I'))
				{
					if(PrintCombo(&pBuff, &nMSS, &Trans->k, &Trans->l, g_IOPorts, g_Input, Trans->i, GetPortID(GetOptIn(Trans->i)), g_frmMgmt))
						return pBuff-pBuff0;
					Trans->j+=2;
				}
				else if(OP_Chk(g_MgmtHtmlLine+Trans->j,'J'))
				{
					switch(GetPortIO(GetOptIn(Trans->i)))
					{
					case OUTPUT:
					case INPUT:
					case MEMORY:
					case MEMORY_SW:
						if(PrintCombo(&pBuff, &nMSS, &Trans->k, &Trans->l, g_BinV, g_InVal, Trans->i, GetComboOpt(GetOptInVal(Trans->i),g_OptsBinV,Trans->Buff), g_frmMgmt))
							return pBuff-pBuff0;
						break;
					case INPUT_ADC:
					case OUTPUT_PWM:
						if(PrintCombo(&pBuff, &nMSS, &Trans->k, &Trans->l, g_AdcV, g_InVal, Trans->i, GetComboOpt(GetOptInVal(Trans->i),g_OptsAdcV,Trans->Buff), g_frmMgmt))
							return pBuff-pBuff0;
						break;
					default:
						if(strcopy_P(&pBuff, &nMSS, &Trans->k, g_HtmlEmpty))
							return pBuff-pBuff0;
					}
					Trans->j+=2;
				}
				else if(OP_Chk(g_MgmtHtmlLine+Trans->j,'X'))
				{
					switch(GetPortIO(GetOptIn(Trans->i)))
					{
					case OUTPUT:
					case INPUT:
					case MEMORY:
					case MEMORY_SW:
						if(GetOptInVal(Trans->i) < AND)
						{
							int Aux = GetOptAux(Trans->i);
							if(Aux == 0xFFFF)
								Aux = 0;
							if(strprint(&pBuff, &nMSS, &Trans->k, g_MgmtHtmlEB,g_InAux,Trans->i,Aux))
								return pBuff-pBuff0;
						}
						else
						{
							if(PrintCombo(&pBuff, &nMSS, &Trans->k, &Trans->l, g_IOPorts2, g_InAux, Trans->i, GetPortID(GetOptAux(Trans->i)), g_frmMgmt))
								return pBuff-pBuff0;
						}
						break;
					case INPUT_ADC:
					case OUTPUT_PWM:
						if(GetOptInVal(Trans->i) < EQ_CHN)
						{
							int Aux = GetOptAux(Trans->i);
							if(Aux == 0xFFFF)
								Aux = 0;
							if(strprint(&pBuff, &nMSS, &Trans->k, g_MgmtHtmlEB,g_InAux,Trans->i,Aux))
								return pBuff-pBuff0;
						}
						else
						{
							if(PrintCombo(&pBuff, &nMSS, &Trans->k, &Trans->l, g_APorts, g_InAux, Trans->i, GetPortID(GetOptAux(Trans->i)), g_frmMgmt))
								return pBuff-pBuff0;
						}
						break;
					default:
						if(strcopy_P(&pBuff, &nMSS, &Trans->k, g_HtmlEmpty))
							return pBuff-pBuff0;
					}
					Trans->j+=2;
				}
				else if(OP_Chk(g_MgmtHtmlLine+Trans->j,'O'))
				{
					if(PrintCombo(&pBuff, &nMSS, &Trans->k, &Trans->l, g_OPorts, g_Output, Trans->i, GetPortID(GetOptOut(Trans->i)), g_frmMgmt))
						return pBuff-pBuff0;
					Trans->j+=2;
				}
				else if(OP_Chk(g_MgmtHtmlLine+Trans->j,'P'))
				{
					switch(GetPortIO(GetOptOut(Trans->i)))
					{
					case OUTPUT:
					case MEMORY:
					case MEMORY_SW:
						if(PrintCombo(&pBuff, &nMSS, &Trans->k, &Trans->l, g_BitV, g_OutVal, Trans->i, GetComboOpt(GetOptOutVal(Trans->i),g_OptsBitV,Trans->Buff), g_frmMgmt))
							return pBuff-pBuff0;
						break;
					case OUTPUT_PWM:
						if(strprint(&pBuff, &nMSS, &Trans->k, g_MgmtHtmlEB, g_OutVal, Trans->i, GetOptOutVal(Trans->i)))
							return pBuff-pBuff0;
						break;
					default:
						if(strcopy_P(&pBuff, &nMSS, &Trans->k, g_HtmlEmpty))
							return pBuff-pBuff0;
					}
					Trans->j+=2;
				}
				else
				{
					*pBuff++ = pgm_read_byte(g_MgmtHtmlLine+Trans->j);
					Trans->j+=1;
					if(!--nMSS)
						return pBuff-pBuff0;
				}
			}
			Trans->j = 0;
			Trans->i++;
		}
		Trans->i = 0;
		Trans->h = 4;
	}

	if(Trans->h == 4)
	{
		if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_MgmtHtmlFormEnd))
			return pBuff-pBuff0;
		Trans->h = 5;
	}

	if(Trans->h == 5)
	{
		if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_HtmlFooter))
			return pBuff-pBuff0;
		Trans->h = -1;
	}

	return pBuff-pBuff0;
}

// Gui
static const prog_char g_CtrlHtmlTable[] =
				"<h1>Controll Pannel</h1>\r\n"
				"<form name='frmCtrl' method='POST' action='/Ctrl'>\r\n"
				"<table border='0' cellpadding='0' cellspacing='0'><tr>\r\n"
				"<td valign='top'>\r\n"
				"<table border='1'><colgroup><col width='140'><col width='40'></colgroup>\r\n";

static const prog_char g_CtrlHtmlLineTop[] =
				"<td>%S</td><td>%S</td>\r\n";

static const prog_char g_CtrlHtmlLine[] =
				"<tr><td class='s%C'>%L</td><td height='29' class='s%C'>%X</td></tr>\r\n";

static const prog_char g_CtrlHtmlSwitch[] =
				"<input type='radio' name='Value%d' value='%S' %S onclick='frmCtrl.submit();'>%S";
static const prog_char g_CtrlHtmlChecked[] =
				"checked='checked'";
//static const prog_chrt g_CtrlHtmlDisabled[] =
//				"disabled='disabled'";

static const prog_char g_CtrlHtmlEB[] =
				"<input name='%s%d' onblur = 'frmCtrl.submit();' value='%d' maxlength='10' size='5' type='text'>\r\n";

static const prog_char g_CtrlHtmlSpacer[] =
				"</table></td><td valign='top'>&nbsp;</td><td valign='top'>\r\n"
				"<table border='1'><colgroup><col width='140'><col width='80'></colgroup>\r\n";


static const prog_char g_CtrlHtmlTableEnd[] =
				"</table></form></td>\r\n"
				"</tr></table>\r\n"
				"<p><a href='/Mgmt'>Management</a>&nbsp;<a href='/Cfg'>Configuration</a>&nbsp;<a href='/Lcd'>Lcd</a></p>";

extern const char g_Input[];
extern const char g_Output[];
extern const char g_Value[];
static const prog_char g_Memory[] = "Memory";
extern const prog_char g_Hi[];
extern const prog_char g_Lo[];

int CtrlHttpData(char* pBuff, int nMSS, tTrans* Trans)
{
	char* pBuff0 = pBuff;

	if(Trans->h == 0)
	{
		if(Trans->j == 0)
		{
			if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_HtmlHeaderBegin))
				return pBuff-pBuff0;
			Trans->j = 1;
		}

		if(Trans->j == 1)
		{
			if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_HtmlHeaderStyle))
				return pBuff-pBuff0;
			Trans->j = 2;
		}

		if(Trans->j == 2)
		{
			if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_HtmlHeaderEnd))
				return pBuff-pBuff0;
			Trans->j = 0;
		}
		Trans->h = 1;
	}

	if(Trans->h == 1)
	{
		if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_CtrlHtmlTable))
			return pBuff-pBuff0;
		Trans->h = 2;
	}

	const int CtrlHtmlLineLen = strlen_P(g_CtrlHtmlLine);

	if(Trans->h == 2)
	{
		if(strprint(&pBuff, &nMSS, &Trans->i, g_CtrlHtmlLineTop,g_Input,g_Value))
			return pBuff-pBuff0;
		Trans->h = 3;
	}

	const int HWPortCount= GetHWPortCount();

	if(Trans->h == 3)
	{
		while(Trans->i<HWPortCount)
		{
			uint8_t IO = GetPortIO(Trans->i);
			if((IO != INPUT && IO != INPUT_ADC) || (IO >= LC_RS && IO <= LC_SR))
			{
				Trans->i++;
				continue;
			}

			while(Trans->j<CtrlHtmlLineLen)
			{
				if(OP_Chk(g_CtrlHtmlLine+Trans->j,'L'))
				{
					char* Label = GetPortLabel(Trans->i,1);
					if(Label)
					{
						if(strcopy_C(&pBuff, &nMSS, &Trans->k, Label))
							return pBuff-pBuff0;
					}
					Trans->j+=2;
				}
				else if(OP_Chk(g_CtrlHtmlLine+Trans->j,'X'))
				{
					switch(GetPortIO(Trans->i))
					{
					case INPUT:
						if(strcopy_P(&pBuff, &nMSS, &Trans->k, GetBit(Trans->i) ? g_Hi : g_Lo))
							return pBuff-pBuff0;
						break;
					case INPUT_ADC:
						itoa(GetVal(Trans->i), g_Buffer, 10);
						if(strcopy_C(&pBuff, &nMSS, &Trans->k, g_Buffer))
							return pBuff-pBuff0;
						break;
					}
					Trans->j+=2;
				}
				else if(OP_Chk(g_CtrlHtmlLine+Trans->j,'C'))
				{
					*pBuff++ = '2';
					Trans->j+=2;
					if(!--nMSS)
						return pBuff-pBuff0;
				}
				else
				{
					*pBuff++ = pgm_read_byte(g_CtrlHtmlLine+Trans->j);
					Trans->j+=1;
					if(!--nMSS)
						return pBuff-pBuff0;
				}
			}
			Trans->j = 0;
			Trans->i++;
		}

		Trans->i = 0;
		Trans->h = 4;
	}

	if(Trans->h == 4)
	{
		if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_CtrlHtmlSpacer))
			return pBuff-pBuff0;
		Trans->h = 5;
	}

	if(Trans->h == 5)
	{
		if(strprint(&pBuff, &nMSS, &Trans->i, g_CtrlHtmlLineTop,g_Output,g_Value))
			return pBuff-pBuff0;
		Trans->h = 6;
	}

	if(Trans->h == 6)
	{
		while(Trans->i<HWPortCount)
		{
			uint8_t IO = GetPortIO(Trans->i);
			if((IO != OUTPUT && IO != OUTPUT_PWM) || (IO >= LC_RS && IO <= LC_SR))
			{
				Trans->i++;
				continue;
			}

			while(Trans->j<CtrlHtmlLineLen)
			{
				if(OP_Chk(g_CtrlHtmlLine+Trans->j,'L'))
				{
					char* Label = GetPortLabel(Trans->i,1);
					if(Label)
					{
						if(strcopy_C(&pBuff, &nMSS, &Trans->k, Label))
							return pBuff-pBuff0;
					}
					Trans->j+=2;
				}
				else if(OP_Chk(g_CtrlHtmlLine+Trans->j,'X'))
				{
					switch(GetPortIO(Trans->i))
					{
					case OUTPUT:
						if(Trans->l == 0)
						{
							if(strprint(&pBuff, &nMSS, &Trans->k, g_CtrlHtmlSwitch,Trans->i, g_Lo, GetBit(Trans->i) ? g_NULL : g_CtrlHtmlChecked, g_Lo))
								return pBuff-pBuff0;
							Trans->l = 1;
						}
						if(Trans->l == 1)
						{
							if(strprint(&pBuff, &nMSS, &Trans->k, g_CtrlHtmlSwitch,Trans->i, g_Hi, GetBit(Trans->i) ? g_CtrlHtmlChecked : g_NULL, g_Hi))
								return pBuff-pBuff0;
							Trans->l = 0;
						}
						break;
					case OUTPUT_PWM:
						if(strprint(&pBuff, &nMSS, &Trans->k, g_CtrlHtmlEB,g_Value,Trans->i,GetVal(Trans->i)))
							return pBuff-pBuff0;
						break;
					}
					Trans->j+=2;
				}
				else if(OP_Chk(g_CtrlHtmlLine+Trans->j,'C'))
				{
					*pBuff++ = '1';
					Trans->j+=2;
					if(!--nMSS)
						return pBuff-pBuff0;
				}
				else
				{
					*pBuff++ = pgm_read_byte(g_CtrlHtmlLine+Trans->j);
					Trans->j+=1;
					if(!--nMSS)
						return pBuff-pBuff0;
				}
			}
			Trans->j = 0;
			Trans->i++;
		}
		Trans->i = 0;
		Trans->h = 7;
	}

	const int MEMPortCount = GetPortCount()-HWPortCount;
	if(Trans->h == 7)
	{
		uint8_t Mem = 0;
		int i;
		for(i=0; i<MEMPortCount;i++)
		{
			if(GetPortIO(HWPortCount+i) == MEMORY_SW)
			{
				Mem = 1;
				break;
			}
		}

		if(Mem)
			Trans->h = 8;
		else
			Trans->h = 11;
	}

	if(Trans->h == 8)
	{
		if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_CtrlHtmlSpacer))
			return pBuff-pBuff0;
		Trans->h = 9;
	}

	if(Trans->h == 9)
	{
		if(strprint(&pBuff, &nMSS, &Trans->i, g_CtrlHtmlLineTop,g_Memory,g_Value))
			return pBuff-pBuff0;
		Trans->h = 10;
	}

	if(Trans->h == 10)
	{
		while(Trans->i<MEMPortCount)
		{
			if(GetPortIO(Trans->i+HWPortCount) != MEMORY_SW)
			{
				Trans->i++;
				continue;
			}

			while(Trans->j<CtrlHtmlLineLen)
			{
				if(OP_Chk(g_CtrlHtmlLine+Trans->j,'L'))
				{
					char* Label = GetPortLabel(Trans->i+HWPortCount,1);
					if(Label)
					{
						if(strcopy_C(&pBuff, &nMSS, &Trans->k, Label))
							return pBuff-pBuff0;
					}
					Trans->j+=2;
				}
				else if(OP_Chk(g_CtrlHtmlLine+Trans->j,'X'))
				{
					if(Trans->l == 0)
					{
						if(strprint(&pBuff, &nMSS, &Trans->k, g_CtrlHtmlSwitch,Trans->i+HWPortCount, g_Lo, GetBit(Trans->i+HWPortCount) ? g_NULL : g_CtrlHtmlChecked, g_Lo))
							return pBuff-pBuff0;
						Trans->l = 1;
					}
					if(Trans->l == 1)
					{
						if(strprint(&pBuff, &nMSS, &Trans->k, g_CtrlHtmlSwitch,Trans->i+HWPortCount, g_Hi, GetBit(Trans->i+HWPortCount) ? g_CtrlHtmlChecked : g_NULL, g_Hi))
							return pBuff-pBuff0;
						Trans->l = 0;
					}
					Trans->j+=2;
				}
				else if(OP_Chk(g_CtrlHtmlLine+Trans->j,'C'))
				{
					*pBuff++ = '5';
					Trans->j+=2;
					if(!--nMSS)
						return pBuff-pBuff0;
				}
				else
				{
					*pBuff++ = pgm_read_byte(g_CtrlHtmlLine+Trans->j);
					Trans->j+=1;
					if(!--nMSS)
						return pBuff-pBuff0;
				}
			}
			Trans->j = 0;
			Trans->i++;
		}
		Trans->i = 0;
		Trans->h = 11;
	}


	if(Trans->h == 11)
	{
		if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_CtrlHtmlTableEnd))
			return pBuff-pBuff0;
		Trans->h = 12;
	}

	if(Trans->h == 12)
	{
		if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_HtmlFooter))
			return pBuff-pBuff0;
		Trans->h = -1;
	}

	return pBuff-pBuff0;
}

// LCD Options
static const prog_char g_LcdHtmlTable[] =
				"<h1>LCD</h1>\r\n"
				"<form form name='frmLcd' method='POST' action='/Lcd'>\r\n"
				"<table border='1'>\r\n";

static const prog_char g_LcdHtmlLine[] =
				"<tr>%S<td>%X</td><td><input value='T' type='submit' name='UpPG%N'><input value='K' type='submit' name='RemPG%N'><input value='B' type='submit' name='DownPG%N'></td></tr>\r\n";

static const prog_char g_LcdHtmlScr[] =
				"<td><input name='%s%c%d' value='%s' maxlength='40' size='20' type='text'></td>";

static const prog_char g_LcdHtmlAdd[] =
				"<tr><td colspan='%d'>Auto Switch Screans Delay: &nbsp;<input name='Auto' onblur = 'frmLcd.submit();' value='%d' maxlength='10' size='5' type='text'></td>";

static const prog_char g_LcdHtmlTableEnd[] =
				"<td><input value='N' type='submit' name='AddPG'></td><tr>\r\n</table>\r\n"
				"<br><input name='Update' type='submit' value='Update'>&nbsp;<input name='Save2' type='submit' value='Save'></form>\r\n";

extern const char g_AddPG[];
extern const char g_UpPG[];
extern const char g_DownPG[];
extern const char g_DownPG[];

extern const char g_Page[];

extern const char g_Action[];
extern const char g_Action[];

int LcdHttpData(char* pBuff, int nMSS, tTrans* Trans)
{
	char* pBuff0 = pBuff;

	if(Trans->h == 0)
	{
		if(Trans->j == 0)
		{
			if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_HtmlHeaderBegin))
				return pBuff-pBuff0;
			Trans->j = 1;
		}

		if(Trans->j ==1)
		{
			if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_HtmlHeaderScript))
				return pBuff-pBuff0;
			Trans->j = 2;
		}

		if(Trans->j == 2)
		{
			if(PrintPortsList(&pBuff, &nMSS, &Trans->k, &Trans->l, &Trans->m, g_OPorts, 5))
				return pBuff-pBuff0;
			Trans->j = 3;
		}

		if(Trans->j == 3)
		{
			if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_HtmlHeaderEnd))
				return pBuff-pBuff0;
			Trans->j = 0;
		}
		Trans->h = 1;
	}

	if(Trans->h == 1)
	{
		if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_LcdHtmlTable))
			return pBuff-pBuff0;
		Trans->h = 2;
	}

	const int LineCount = GetLCDLines();
	const int ScrCount = GetScrCount();
	if(Trans->h == 2)
	{
		const int LcdHtmlLineLen = strlen_P(g_LcdHtmlLine);

		while(Trans->i<ScrCount)
		{
			while(Trans->j<LcdHtmlLineLen)
			{
				if(OP_Chk(g_LcdHtmlLine+Trans->j,'N'))
				{
					itoa(Trans->i, g_Buffer, 10);
					if(strcopy_C(&pBuff, &nMSS, &Trans->k, g_Buffer))
						return pBuff-pBuff0;
					Trans->j+=2;
				}
				else if(OP_Chk(g_LcdHtmlLine+Trans->j,'S'))
				{
					while(Trans->l<LineCount)
					{
						if(strprint(&pBuff, &nMSS, &Trans->k, g_LcdHtmlScr,g_Page,'A'+Trans->l,Trans->i,GetScr(Trans->i,Trans->l)))
							return pBuff-pBuff0;
						Trans->l++;
					}
					Trans->l = 0;
					Trans->j+=2;
				}
				else if(OP_Chk(g_LcdHtmlLine+Trans->j,'X'))
				{
					if(PrintCombo(&pBuff, &nMSS, &Trans->k, &Trans->l, g_OPorts, g_Action, Trans->i, GetPortID(GetScrAction(Trans->i)), g_NULL))
						return pBuff-pBuff0;
					Trans->j+=2;
				}
				else
				{
					*pBuff++ = pgm_read_byte(g_LcdHtmlLine+Trans->j);
					Trans->j+=1;
					if(!--nMSS)
						return pBuff-pBuff0;
				}
			}
			Trans->j = 0;
			Trans->i++;
		}
		Trans->i = 0;
		Trans->h = 3;
	}

	if(Trans->h == 3)
	{
		if(strprint(&pBuff, &nMSS, &Trans->i, g_LcdHtmlAdd, LineCount+1, GetLCDAuto()))
			return pBuff-pBuff0;
		Trans->h = 4;
	}

	if(Trans->h == 4)
	{
		if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_LcdHtmlTableEnd))
			return pBuff-pBuff0;
		Trans->h = 5;
	}

	if(Trans->h == 5)
	{
		if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_HtmlFooter))
			return pBuff-pBuff0;
		Trans->h = -1;
	}

	return pBuff-pBuff0;
}

int MakeHttpData(char* pBuff, int nMSS, tTrans* Trans)
{
	if(Trans->Index == 0)
		return CtrlHttpData(pBuff, nMSS, Trans);
	else if(Trans->Index == 1)
		return CfgHttpData(pBuff, nMSS, Trans);
	else if(Trans->Index == 2)
		return MgmtHttpData(pBuff, nMSS, Trans);
	else if(Trans->Index == 3)
		return LcdHttpData(pBuff, nMSS, Trans);
	return 0;
}

// Request Processing
uint8_t strfind_P(char** pBuff, int *iLen, int* iPos, const prog_char* pStr)
{
	while(*iLen)
	{
		(*iLen)--;
		if(**pBuff == pgm_read_byte(pStr+*iPos))
		{
			(*iPos)++;
			if(pgm_read_byte(pStr+*iPos) == '\0')
			{
				(*pBuff)++;
				*iPos = 0;
				return 0;
			}
		}
		else
			*iPos = 0;
		(*pBuff)++;
	}
	return 1;
}

uint8_t strcopyto_P(char* pStr, int* iPos, char** pBuff, int *iLen, prog_char* pEnd, int *iEnd, int *iLen2)
{
	while(*iLen)
	{
		(*iLen)--;
		if(iLen2) (*iLen2)--;
		pStr[*iPos] = **pBuff;
		if(**pBuff == pgm_read_byte(pEnd+*iEnd))
		{
			(*iEnd)++;
			if(pgm_read_byte(pEnd+*iEnd) == '\0')
			{
				*iEnd = 0;
				(*pBuff)++;
				return 0;
			}
		}
		else
			*iEnd = 0;
		(*pBuff)++;
		(*iPos)++;
	}
	if(iLen2 && *iLen2 == 0)
		return 0;
	return 1;
}

int ParseHeader(char* pBuff, int iLen, tTrans* Trans)
{
	if(Trans->h == 0)
	{
		if(Trans->m < 4)
		{
			int togo = iLen <= (4-Trans->m) ? iLen : (4-Trans->m);
			strncpy(Trans->Buff+Trans->m,pBuff,togo);
			Trans->m += togo;
			pBuff += togo;
			iLen -= togo;
			if(Trans->m < 4)
				return 1;
		}

		if(strncmp_P(Trans->Buff,PSTR("GET"),3) == 0)
		{
			Trans->m = 0;
			Trans->h = 2;
		}
		else if(strncmp_P(Trans->Buff,PSTR("POST"),4) == 0)
		{
			if(iLen == 0)
				return 1;
			pBuff += 1;
			iLen -= 1; // skip the space
			Trans->m = 0;
			Trans->h = 1;
		}
		else
		{
			Trans->m = 0;
			Trans->h = 10;
			return 1;
		}
	}

	if(Trans->h == 1 || Trans->h == 2)
	{
		if(strcopyto_P(Trans->Buff, &Trans->m, &pBuff, &iLen, PSTR("HTTP"), &Trans->l, NULL))
			return 1;
		Trans->m = 0;

		if(strncmp_P(Trans->Buff,PSTR("/ "),2) == 0 || strncmp_P(Trans->Buff,PSTR("/Ctrl "),6) == 0)
		{
			Trans->Index = 0;
			Trans->h += 2;
		}
		else if(strncmp_P(Trans->Buff,PSTR("/Cfg "),5) == 0)
		{
			Trans->Index = 1;
			Trans->h += 2;
		}
		else if(strncmp_P(Trans->Buff,PSTR("/Mgmt "),6) == 0)
		{
			Trans->Index = 2;
			Trans->h += 2;
		}
		else if(strncmp_P(Trans->Buff,PSTR("/Lcd "),5) == 0)
		{
			Trans->Index = 3;
			Trans->h += 2;
		}
		else
		{
			Trans->h = 10;
			return 1;
		}
	}

	if(Trans->h == 3)
	{

		if(Trans->k == 0)
		{
			if(strfind_P(&pBuff, &iLen, &Trans->m,PSTR("Content-Length:")))
				return 1;
			Trans->k = 1;
		}

		if(Trans->k == 1)
		{
			if(strcopyto_P(Trans->Buff, &Trans->m, &pBuff, &iLen, PSTR("\r\n"), &Trans->l, NULL))
				return 1;

			Trans->Buff[Trans->m] = '\0';
			Trans->m = 2; // mark that we already have a \r\n pair!
			Trans->j = atoi(Trans->Buff);

			Trans->k = 2;
		}

		if(Trans->k == 2)
		{
			if(strfind_P(&pBuff, &iLen, &Trans->m,PSTR("\r\n\r\n")))
				return 1;
			Trans->k = 3;
		}

		if(Trans->k == 3)
		{
			while(Trans->j)
			{
				if(strcopyto_P(Trans->Buff, &Trans->m, &pBuff, &iLen, PSTR("&"), &Trans->l, &Trans->j))
					return 1;
				Trans->Buff[Trans->m] = '\0';
				Trans->m = 0;
				shell_input(Trans->Buff, 0);
			}
			Trans->k = 0;
		}
		shell_queue();
		Trans->h = 0;
	}
	else if(Trans->h == 4)
		Trans->h = 0;

	return 0;
}

