/*
 * Ports.c
 *
 *      Author: Trax
 */

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <avr/io.h>
#include "Ports.h"

tGpio g_Gpio[] =
{
	{&PINC, &PORTC, &DDRC, PC0, NONE, NONE, "C0", NULL, OUTPUT},
	{&PINC, &PORTC, &DDRC, PC1, NONE, NONE, "C1", NULL, OUTPUT},
	{&PINC, &PORTC, &DDRC, PC2, NONE, NONE, "C2", NULL, OUTPUT},
	{&PINC, &PORTC, &DDRC, PC3, NONE, NONE, "C3", NULL, OUTPUT},
	{&PINC, &PORTC, &DDRC, PC4, NONE, NONE, "C4", NULL, OUTPUT},
	{&PINC, &PORTC, &DDRC, PC5, NONE, NONE, "C5", NULL, OUTPUT},
	{&PINC, &PORTC, &DDRC, PC6, NONE, NONE, "C6", NULL, OUTPUT},
	{&PINC, &PORTC, &DDRC, PC7, NONE, NONE, "C7", NULL, OUTPUT},

	{&PINA, &PORTA, &DDRA, PA0, 0, NONE, "A0", NULL, INPUT},
	{&PINA, &PORTA, &DDRA, PA1, 1, NONE, "A1", NULL, INPUT},
	{&PINA, &PORTA, &DDRA, PA2, 2, NONE, "A2", NULL, INPUT},
	{&PINA, &PORTA, &DDRA, PA3, 3, NONE, "A3", NULL, INPUT},
	{&PINA, &PORTA, &DDRA, PA4, 4, NONE, "A4", NULL, INPUT_ADC},
	{&PINA, &PORTA, &DDRA, PA5, 5, NONE, "A5", NULL, INPUT_ADC},
	{&PINA, &PORTA, &DDRA, PA6, 6, NONE, "A6", NULL, INPUT_ADC},
	{&PINA, &PORTA, &DDRA, PA7, 7, NONE, "A7", NULL, INPUT_ADC},

	{&PIND, &PORTD, &DDRD, PD2, NONE, NONE, "D2", NULL, LC_RS},	// RS
	{&PIND, &PORTD, &DDRD, PD3, NONE, NONE, "D3", NULL, LC_EN},	// EN
	{&PIND, &PORTD, &DDRD, PD4, NONE, 0x1B, "D4", NULL, LC_D4},	// D4
	{&PIND, &PORTD, &DDRD, PD5, NONE, 0x1A, "D5", NULL, LC_D5},	// D5
	{&PIND, &PORTD, &DDRD, PD6, NONE, 0x2B, "D6", NULL, LC_D6},	// D6
	{&PIND, &PORTD, &DDRD, PD7, NONE, 0x2A, "D7", NULL, LC_D7},	// D7
	{&PINB, &PORTB, &DDRB, PB0, NONE, NONE, "B0", NULL, LC_SL},	// SL
	{&PINB, &PORTB, &DDRB, PB3, NONE, NONE, "B3", NULL, LC_SR},	// SR

	{&PINB, &PORTB, &DDRB, PB1, NONE, NONE, "B1", NULL, NONE}
};

tVPort g_VPort[] =
{
	{0, "M0", NULL, NONE},
	{0, "M1", NULL, NONE},
	{0, "M2", NULL, NONE},
	{0, "M3", NULL, NONE},
	{0, "M4", NULL, NONE},
	{0, "M5", NULL, NONE},
	{0, "M6", NULL, NONE},
	{0, "M7", NULL, NONE},
	{0, "N0", NULL, NONE},
	{0, "N1", NULL, NONE},
	{0, "N2", NULL, NONE},
	{0, "N3", NULL, NONE},
	{0, "N4", NULL, NONE},
	{0, "N5", NULL, NONE},
	{0, "N6", NULL, NONE},
	{0, "N7", NULL, NONE}
};

uint8_t GetPortCount()
{
	return ARRSIZE(g_Gpio)+ARRSIZE(g_VPort);
}

uint8_t GetHWPortCount()
{
	return ARRSIZE(g_Gpio);
}

uint8_t GetPortADC(uint8_t Port)
{
	if(Port >= ARRSIZE(g_Gpio))
		return NONE;
	return g_Gpio[Port].nADC;
}

uint8_t GetPortPWM(uint8_t Port)
{
	if(Port >= ARRSIZE(g_Gpio))
		return NONE;
	return g_Gpio[Port].nPWM;
}

uint8_t GetPortMEM(uint8_t Port)
{
	if(Port >= ARRSIZE(g_Gpio))
	{
		Port -= ARRSIZE(g_Gpio);
		if(Port < ARRSIZE(g_VPort))
			return Port;
	}
	return NONE;
}

char* GetPortID(uint8_t Port)
{
	if(Port >= ARRSIZE(g_Gpio))
	{
		Port -= ARRSIZE(g_Gpio);
		if(Port >= ARRSIZE(g_VPort))
			return "";
		else
			return g_VPort[Port].ID;
	}
	return g_Gpio[Port].ID;
}

char* GetPortLabel(uint8_t Port, uint8_t UseID)
{
	if(Port >= ARRSIZE(g_Gpio))
	{
		Port -= ARRSIZE(g_Gpio);
		if(Port >= ARRSIZE(g_VPort))
			return NULL;
		return (UseID && !g_VPort[Port].Label) ? g_VPort[Port].ID : g_VPort[Port].Label;
	}
	return (UseID && !g_Gpio[Port].Label) ? g_Gpio[Port].ID : g_Gpio[Port].Label;
}

void SetPortLabel(uint8_t Port, const char* NewLabel)
{
	char** Label = Port >= ARRSIZE(g_Gpio) ? &g_VPort[Port-ARRSIZE(g_Gpio)].Label : &g_Gpio[Port].Label;
	if(*Label)
	{
		free(*Label);
		*Label = NULL;
	}
	int Len = strlen(NewLabel);
	if(Len)
	{
		*Label = (char*)malloc(Len+1);
		strcpy(*Label,NewLabel);
	}
}

uint8_t GetPortIO(uint8_t Port)
{
	if(Port == -1)
		return NONE;
	if(Port >= ARRSIZE(g_Gpio))
	{
		Port -= ARRSIZE(g_Gpio);
		if(Port >= ARRSIZE(g_VPort))
			return NONE;
		return g_VPort[Port].IO;
	}
	return g_Gpio[Port].IO;
}

void SetPortIO(uint8_t Port, uint8_t IO)
{
	if(Port >= ARRSIZE(g_Gpio))
	{
		Port -= ARRSIZE(g_Gpio);
		if(Port >= ARRSIZE(g_VPort))
			return;
		g_VPort[Port].IO = IO;
	}
	else
		g_Gpio[Port].IO = IO;
}

uint8_t GetPort(const char *pID)
{
	const int PortCount = GetPortCount();
	int i;
	for(i=0;i<PortCount;i++)
	{
		if(strcmp(GetPortID(i),pID) == 0)
			return i;
	}
	return -1;
}

void SetBitIO(uint8_t Port, uint8_t O)
{
	if(O)
		*(g_Gpio[Port].pDdr) |= (1<<g_Gpio[Port].nBit);
	else
		*(g_Gpio[Port].pDdr) &= ~(1<<g_Gpio[Port].nBit);
}

uint8_t GetBit(uint8_t Port)
{
	if(Port >= ARRSIZE(g_Gpio))
	{
		Port -= ARRSIZE(g_Gpio);
		if(Port >= ARRSIZE(g_VPort))
			return 0;
		return g_VPort[Port].nBit;
	}
	else
	{
		if((*g_Gpio[Port].pDdr >> g_Gpio[Port].nBit) & 1)
			return (*g_Gpio[Port].pPort >> g_Gpio[Port].nBit) & 1;
		else
			return (*g_Gpio[Port].pPin >> g_Gpio[Port].nBit) & 1;
	}
}

void SetBit(uint8_t Port, uint8_t Bit)
{
	if(Port >= ARRSIZE(g_Gpio))
	{
		Port -= ARRSIZE(g_Gpio);
		if(Port >= ARRSIZE(g_VPort))
			return;
		g_VPort[Port].nBit = Bit;
	}
	else
	{
		if(!(*g_Gpio[Port].pDdr >> g_Gpio[Port].nBit) & 1)
			*(g_Gpio[Port].pDdr) |= (1<<g_Gpio[Port].nBit);

		if(Bit)
			*g_Gpio[Port].pPort |= (1<<g_Gpio[Port].nBit);
		else
			*g_Gpio[Port].pPort &= ~(1<<g_Gpio[Port].nBit);
	}
}

uint16_t ReadChannel(uint8_t mux)
{
	uint8_t i;
	uint16_t result;

	ADMUX = mux;                      // Kanal waehlen
	ADMUX |= (0<<REFS1) | (0<<REFS0); // externe Referenzspannung nutzen //00 Externes AREF //01 AVCC als Referenz // 11 Interne 2,56 Volt

	ADCSRA = (1<<ADEN) | (1<<ADPS1) | (1<<ADPS0);    // Frequenzvorteiler
							   // setzen auf 8 (1) und ADC aktivieren (1)

	/* nach Aktivieren des ADC wird ein "Dummy-Readout" empfohlen, man liest
	 also einen Wert und verwirft diesen, um den ADC "warmlaufen zu lassen" */
	ADCSRA |= (1<<ADSC);              // eine ADC-Wandlung
	while ( ADCSRA & (1<<ADSC) ) {
		;     // auf Abschluss der Konvertierung warten
	}
	result = ADCW;  // ADCW muss einmal gelesen werden,
				  // sonst wird Ergebnis der nchsten Wandlung
				  // nicht bernommen.

	/* Eigentliche Messung - Mittelwert aus 4 aufeinanderfolgenden Wandlungen */
	result = 0;
	for( i=0; i<4; i++ )
	{
		ADCSRA |= (1<<ADSC);            // eine Wandlung "single conversion"
		while ( ADCSRA & (1<<ADSC) ) {
			;   // auf Abschluss der Konvertierung warten
		}
		result += ADCW;		    // Wandlungsergebnisse aufaddieren
	}
	ADCSRA &= ~(1<<ADEN);             // ADC deaktivieren (2)

	result /= 4;                     // Summe durch vier teilen = arithm. Mittelwert

	return result;
}

uint16_t GetVal(uint8_t Port)
{
	if(Port >= ARRSIZE(g_Gpio) || g_Gpio[Port].nADC == NONE)
		return 0;
	return ReadChannel(g_Gpio[Port].nADC);
}

void SetVal(uint8_t Port, uint16_t Val)
{
	if(!(*g_Gpio[Port].pDdr >> g_Gpio[Port].nBit) & 1)
		*(g_Gpio[Port].pDdr) |= (1<<g_Gpio[Port].nBit);

	if(g_Gpio[Port].nPWM == 0x1A) //OC1A // PD5
	{
		if(Val)
		{
			TCCR1A = (1<<WGM11) | (1<<WGM10) | (1<<COM1A1);
			TCCR1B = (0<<WGM13) | (1<<WGM12) | (1<<CS10) | (0<<CS11) | (0<<CS12);

			OCR1A = Val;
		}
		else
		{
			TCCR1A &= ~(1<<COM1A1);
			*(g_Gpio[Port].pPort) &= ~(1<<g_Gpio[Port].nBit);
		}
	}
	else if(g_Gpio[Port].nPWM == 0x1B) //OC1B // PD4
	{
		if(Val)
		{
			TCCR1A = (1<<WGM11) | (1<<WGM10) | (1<<COM1B1);
			TCCR1B = (0<<WGM13) | (1<<WGM12) | (1<<CS10) | (0<<CS11) | (0<<CS12);
			OCR1B = Val;
		}
		else
		{
			TCCR1A &= ~(1<<COM1B1);
			*(g_Gpio[Port].pPort) &= ~(1<<g_Gpio[Port].nBit);
		}
	}
	else if(g_Gpio[Port].nPWM == 0x2A) //OC2A // PD7
	{
		if(Val)
		{
			TCCR2A = (0<<WGM11) | (1<<WGM10) | (1<<COM2A1);
			TCCR2B = (0<<WGM13) | (1<<WGM12) | (1<<CS10) | (0<<CS11) | (0<<CS12);
			OCR2A = Val/4;
		}
		else
		{
			TCCR2A &= ~(1<<COM2A1);
			*(g_Gpio[Port].pPort) &= ~(1<<g_Gpio[Port].nBit);
		}
	}
	else if(g_Gpio[Port].nPWM == 0x2B) //OC2B // PD6
	{
		if(Val)
		{
			TCCR2A = (0<<WGM11) | (1<<WGM10) | (1<<COM2B1);
			TCCR2B = (0<<WGM13) | (1<<WGM12) | (1<<CS10) | (0<<CS11) | (0<<CS12);
			OCR2B = Val/4;
		}
		else
		{
			TCCR2A &= ~(1<<COM2B1);
			*(g_Gpio[Port].pPort) &= ~(1<<g_Gpio[Port].nBit);
		}
	}

	// OC0A // PB3
}
