/*
 * Options.c
 *
 *      Author: Trax
 */

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include "Screens.h"
#include "Ports.h"
#include "Hardware/Lcd.h"

extern tGpio g_Gpio[];

tLCDPins g_LCDPins [] = {
	{"RS", 16, 16, LC_RS},
	{"EN", 17, 17, LC_EN},
	{"D4", 18, 18, LC_D4},
	{"D5", 19, 19, LC_D5},
	{"D6", 20, 20, LC_D6},
	{"D7", 21, 21, LC_D7},
	{"SL", 22, 22, LC_SL},
	{"SR", 23, 23, LC_SR}
};

uint8_t g_LCDSupport = 1;
uint8_t g_LCDLines = 2;
uint8_t g_LCDAuto = 0;

uint8_t GetLCDPort(uint8_t Port)
{
	return (g_LCDSupport == 1 ? g_LCDPins[Port].DefPort : g_LCDPins[Port].Port);
}

const char* GetLCDPin(uint8_t Pin)
{
	if(Pin >= 8)
		return NULL;
	return g_LCDPins[Pin].c;
}

char* GetLCDPinID(uint8_t Pin)
{
	if(g_LCDSupport == 0 || Pin >= 8)
		return "";
	return GetPortID(GetLCDPort(Pin));
}

void SetLCDPin(uint8_t Pin, uint8_t Port){
	if(Port >= GetHWPortCount())
		return;
	if(g_LCDSupport == 2)
	{
		g_LCDPins[Pin].Port = Port;
		g_Gpio[Port].IO = Pin+LC_RS;
	}
}

uint8_t GetLCDSupport(){
	return g_LCDSupport;
}
void SetLCDSupport(uint8_t LCDSupport)
{
	if(g_LCDSupport == LCDSupport)
		return;

	int i;
	for(i=0; i<GetHWPortCount();i++)
	{
		int IO = GetPortIO(i);
		if(IO >= LC_RS && IO <= LC_SR)
			SetPortIO(i,NONE);
	}

	switch(LCDSupport)
	{
	case 1:
		for(i=0; i<8;i++)
			SetPortIO(g_LCDPins[i].DefPort,g_LCDPins[i].IO);
		break;
	case 2:
		for(i=0; i<8;i++)
			SetPortIO(g_LCDPins[i].Port,g_LCDPins[i].IO);
		break;
	}

	g_LCDSupport = LCDSupport;
}
uint8_t GetLCDLines(){
	return g_LCDLines;
}
void SetLCDLines(uint8_t Lines){
	if(Lines > 0 && Lines < 5)
		g_LCDLines = Lines;
}
uint16_t GetLCDAuto(){
	return g_LCDAuto;
}
void SetLCDAuto(uint16_t Auto){
	g_LCDAuto = Auto;
}

// Screans
tScreen* g_Screen = NULL;
uint8_t	g_ScreenCnt = 0;

uint8_t GetScrCount(){
	return g_ScreenCnt;
}

char* GetScr(uint8_t i, uint8_t j)
{
	if(i >= g_ScreenCnt || j >= 4)
		return NULL;
	return g_Screen[i].L[j] ? g_Screen[i].L[j] : "";
}

uint8_t GetScrAction(uint8_t i)
{
	if(i >= g_ScreenCnt)
		return NONE;
	return g_Screen[i].Action;
}

void SetScrAction(uint8_t i, uint8_t Action)
{
	if(i >= g_ScreenCnt)
		return;
	g_Screen[i].Action = Action;
}

void SetScr(uint8_t i, uint8_t j, const char* Scr)
{
	if(j >= 4)
		return;

	if(i >= g_ScreenCnt)
	{
		AddScreen();
		i = g_ScreenCnt-1;
	}

	if(g_Screen[i].L[j])
	{
		free(g_Screen[i].L[j]);
		g_Screen[i].L[j] = NULL;
	}
	int Len = strlen(Scr);
	if(Len)
	{
		g_Screen[i].L[j] = (char*)malloc(Len+1);
		strcpy(g_Screen[i].L[j],Scr);
	}
}

void AddScreen()
{
	if(g_ScreenCnt)
		g_Screen = (tScreen*) realloc(g_Screen,sizeof(tScreen)*(g_ScreenCnt+1));
	else
		g_Screen = (tScreen*) malloc(sizeof(tScreen));

	uint8_t j;
	for (j=0;j<4;j++)
		g_Screen[g_ScreenCnt].L[j] = NULL;
	g_Screen[g_ScreenCnt].Action = -1;

	g_ScreenCnt++;
}

void UpScreen(uint8_t i)
{
	if(i >= g_ScreenCnt || i == 0)
		return;

	tScreen temp;
	memcpy(&temp,&g_Screen[i-1],sizeof(tScreen));
	memcpy(&g_Screen[i-1],&g_Screen[i],sizeof(tScreen));
	memcpy(&g_Screen[i],&temp,sizeof(tScreen));
}

void RemoveScreen(uint8_t i)
{
	if(i >= g_ScreenCnt)
		return;

	uint8_t j;
	for (j=0;j<4;j++)
	{
		if(g_Screen[i].L[j])
			free(g_Screen[i].L[j]);
	}

	if(g_ScreenCnt == 1)
	{
		free(g_Screen);
		g_Screen = NULL;
	}
	else
	{
		tScreen* l_Screen = (tScreen*)malloc(sizeof(tScreen)*(g_ScreenCnt-1));

		if(i > 0)
			memcpy(l_Screen,g_Screen,i*sizeof(tScreen));
		if(i+1 < g_ScreenCnt)
			memcpy(&l_Screen[i],&g_Screen[i+1],(g_ScreenCnt-(i+1))*sizeof(tScreen));

		free(g_Screen);
		g_Screen = l_Screen;
	}
	g_ScreenCnt--;
}

void DownScreen(uint8_t i)
{
	if(i >= g_ScreenCnt || i == g_ScreenCnt -1)
		return;

	tScreen temp;
	memcpy(&temp,&g_Screen[i+1],sizeof(tScreen));
	memcpy(&g_Screen[i+1],&g_Screen[i],sizeof(tScreen));
	memcpy(&g_Screen[i],&temp,sizeof(tScreen));
}

//////////////////////////////////////////////////////////////////////
void InitScreen()
{
	if(g_LCDSupport == 1)
		lcd_init(0);
	else if(g_LCDSupport == 2)
		lcd_init(g_LCDPins);
	else
		return;

	if(GetLCDPort(SL) != NONE)
		SetBitIO(GetLCDPort(SL),0);
    if(GetLCDPort(SR) != NONE)
		SetBitIO(GetLCDPort(SR),0);

    lcd_string("NetIO");

    set_cursor(0,2);
    lcd_string("WebSevr");

}

extern char g_Buffer[128];

uint8_t g_LCDIndex = 0xFF;
void SetUserPage(uint8_t Shift)
{
	g_LCDIndex += Shift;
	if(g_LCDIndex < 0)
		g_LCDIndex = GetScrCount()-1;
	else if(g_LCDIndex >= GetScrCount())
		g_LCDIndex = 0;

	lcd_clear();

	uint8_t i;
	for(i=0;i<GetLCDLines();i++)
	{
		set_cursor(0,i+1);

		uint8_t iPos = 0;
		char ID[3];
		char* Line = g_Screen[g_LCDIndex].L[i];
		if(!Line)
			continue;

		while(*Line)
		{
			if(*Line == '$')
			{
				strncpy(ID,Line+1,2);
				ID[2] = 0;
				int Port = GetPort(ID);
				switch(GetPortIO(Port))
				{
				case INPUT:
				case OUTPUT:
				case MEMORY:
				case MEMORY_SW:
					strcpy(g_Buffer+iPos, GetBit(Port) ? "Hi" : "Lo");
					break;
				case INPUT_ADC:
				case OUTPUT_PWM:
					itoa(GetVal(Port), g_Buffer+iPos, 10);
					break;
				default:
					strcpy(g_Buffer+iPos, "n/a");
				}
				iPos = strlen(g_Buffer);
				Line+=3;
			}
			else
			{
				*(g_Buffer+iPos) = *Line;
				iPos++;
				Line++;
			}
		}
		g_Buffer[iPos] = 0;

		lcd_string(g_Buffer);
	}
}

void SetSysPage(int Shift)
{
	uint8_t IO = 0;
	uint8_t i = 0;
	do{
		g_LCDIndex += Shift;
		if(g_LCDIndex < 0)
			g_LCDIndex = GetHWPortCount()-1;
		else if(g_LCDIndex >= GetHWPortCount())
			g_LCDIndex = 0;

		if(i++ > GetHWPortCount())
			return;

		IO = GetPortIO(g_LCDIndex);
	}while(IO != OUTPUT && IO != INPUT && IO != INPUT_ADC && IO != OUTPUT_PWM);

	lcd_clear();

	set_cursor(0,1);
	lcd_string(GetPortLabel(g_LCDIndex,1));

	switch(IO)
	{
	case OUTPUT:
	case INPUT:
		sprintf_P(g_Buffer,PSTR("= %s"),GetBit(g_LCDIndex) ? "Hi" : "Lo");
		break;
	case INPUT_ADC:
	case OUTPUT_PWM:
		sprintf_P(g_Buffer,PSTR("= %i"),GetVal(g_LCDIndex));
		break;
	}
	set_cursor(0,2);
	lcd_string(g_Buffer);
}

uint8_t nSL = 0;
uint8_t nSR = 0;
void CheckSwitch()
{
	if(!nSL && GetLCDPort(SL) != NONE && GetBit(GetLCDPort(SL)))
	{
		nSL = 1;
		if(!(GetLCDPort(SR) != NONE && GetBit(GetLCDPort(SR))))
			nSR = 0;
	}
	if(!nSR &&GetLCDPort(SR) != NONE && GetBit(GetLCDPort(SR)))
	{
		nSR = 1;
		if(!(GetLCDPort(SL) != NONE && GetBit(GetLCDPort(SL))))
			nSL = 0;
	}
}

uint16_t g_LCDCountdown = 0;
void SwitchPage()
{
	uint8_t Shift = 0;
	if(nSL && nSR)
	{
		if(!GetScrCount())
		{
			if(GetPortIO(g_LCDIndex) == OUTPUT)
				SetBit(g_LCDIndex,!GetBit(g_LCDIndex));
		}
		else
		{
			uint16_t Action = GetScrAction(g_LCDIndex);
			if(Action != NONE)
				SetBit(Action,!GetBit(Action));
		}
	}
	else if(nSL)
	{
		Shift = -1;
	}
	else if(nSR)
	{
		Shift = 1;
	}
	else if(g_LCDAuto && g_LCDCountdown-- == 0)
	{
		g_LCDCountdown = g_LCDAuto;
		Shift = 1;
	}

	if(GetScrCount())
		SetUserPage(Shift);
	else
		SetSysPage(Shift);

	nSL = nSR = 0;
}
