/*
 * Settings.c
 *
 *      Author: Trax
 */
#include <stdio.h>
#include <stdlib.h>
#include <avr/io.h>
#include <avr/eeprom.h>
#include <string.h>
#include "Net/uip/uip.h"
#include "settings.h"
#include <avr/pgmspace.h>
#include "ports.h"
#include "options.h"
#include "screens.h"

extern tGpio g_Gpio[];
extern tLCDPins g_LCDPins[];
extern tVPort g_VPort[];
extern uint8_t g_LCDSupport;
extern uint8_t g_LCDLines;
extern uint8_t g_LCDAuto;
extern tOption* g_Option;
extern tScreen* g_Screen;


extern char g_Buffer[128];

void eeprom_read_IP4 (const uint16_t __p, uip_ipaddr_t *__a)
{
	uip_ipaddr(*__a,
			eeprom_read_byte((uint8_t*)__p),
			eeprom_read_byte((uint8_t*)(__p+1)),
			eeprom_read_byte((uint8_t*)(__p+2)),
			eeprom_read_byte((uint8_t*)(__p+3)));
}

void eeprom_write_IP4 (const uint16_t __p, uip_ipaddr_t *__a)
{
	eeprom_write_byte((uint8_t*)__p,uip_ipaddr1(*__a));
	eeprom_write_byte((uint8_t*)(__p+1),uip_ipaddr2(*__a));
	eeprom_write_byte((uint8_t*)(__p+2),uip_ipaddr3(*__a));
	eeprom_write_byte((uint8_t*)(__p+3),uip_ipaddr4(*__a));
}

void Load()
{
	if(eeprom_read_byte((uint8_t*)EE_VER) != CUR_VER)
		return;

	uint8_t i,j,l;
	uint16_t k = 0;
	for(i=0; i<25;i++)
	{
		g_Gpio[i].IO = eeprom_read_byte((uint8_t*)EE_IO+i);
		if(g_Gpio[i].IO >= LC_RS && g_Gpio[i].IO <= LC_SR)
			g_LCDPins[g_Gpio[i].IO-LC_RS].Port = i;

		for (j=0; (g_Buffer[j] = eeprom_read_byte((uint8_t*)EE_LBLS+k++)) != 0 && EE_LBLS+k < EE_END;j++);

		SetPortLabel(i,g_Buffer);
	}

	uint8_t LCD = eeprom_read_byte((uint8_t*)EE_LCD_C);
	g_LCDSupport = LCD & 0x0F;
	g_LCDLines = (LCD & 0xF0) >> 4;
	g_LCDAuto = eeprom_read_byte((uint8_t*)EE_LCD_D);

	for(i=0; i<16;i++)
	{
		g_VPort[i].IO = eeprom_read_byte((uint8_t*)EE_M+i);

		for (j=0; (g_Buffer[j] = eeprom_read_byte((uint8_t*)EE_LBLS+k++)) != 0 && EE_LBLS+k < EE_END;j++);

		SetPortLabel(25+i,g_Buffer);
	}

	uint8_t OptionCnt = eeprom_read_byte((uint8_t*)EE_OPT_C);
	for(i=0; i<OptionCnt;i++)
	{
		for(j=0; j<8;j++)
		{
			g_Buffer[j] = eeprom_read_byte((uint8_t*)(EE_OPTS+(i*8)+j));
		}
		AddOption((uint8_t*)g_Buffer);
	}

	for (i=j=k=l=0; (g_Buffer[l] = eeprom_read_byte((uint8_t*)EE_END-k++)) != 0xFF && EE_END-k > EE_LBLS;)
	{
		if(g_Buffer[l] == 0)
		{
			SetScr(i,j,g_Buffer);
			j++;
			if(j>=4)
			{
				SetScrAction(i, eeprom_read_byte((uint8_t*)EE_END-k++));
				i++;
				j=0;
			}
			l=0;
		}
		else
			l++;
	}
}

void Save(int n)
{
	if(eeprom_read_byte((uint8_t*)EE_VER) != CUR_VER)
	{
		n = 4;
		eeprom_write_byte((uint8_t*)EE_VER, CUR_VER);
	}

	uint8_t i,j,l;
	uint16_t k;
	if(n == 0 || n == 4)
	{
		k=0;
		for(i=0; i<25;i++)
		{
			eeprom_write_byte((uint8_t*)EE_IO+i, g_Gpio[i].IO);

			if(g_Gpio[i].Label)
			{
				for(j=0;g_Gpio[i].Label[j];j++)
				{
					eeprom_write_byte((uint8_t*)(EE_LBLS+k),g_Gpio[i].Label[j]);
					k++;
				}
			}
			eeprom_write_byte((uint8_t*)(EE_LBLS+k),0);
			k++;
		}

		uint8_t LCD = (g_LCDSupport & 0xF) | ((g_LCDLines << 4) & 0xF0);
		eeprom_write_byte((uint8_t*)EE_LCD_C, LCD);
		eeprom_write_byte((uint8_t*)EE_LCD_D, g_LCDAuto);

		for(i=0; i<16;i++)
		{
			eeprom_write_byte((uint8_t*)EE_M+i, g_VPort[i].IO);

			if(g_VPort[i].Label)
			{
				for(j=0;g_VPort[i].Label[j];j++)
				{
					eeprom_write_byte((uint8_t*)(EE_LBLS+k),g_VPort[i].Label[j]);
					k++;
				}
			}
			eeprom_write_byte((uint8_t*)(EE_LBLS+k),0);
			k++;
		}
	}

	if(n == 1 || n == 4)
	{
		uint8_t OptionCnt = GetOptCount();
		if(OptionCnt > 48)
			OptionCnt = 48;
		eeprom_write_byte((uint8_t*)EE_OPT_C, OptionCnt);
		for(i=0; i<OptionCnt;i++)
		{
			memcpy(g_Buffer,&g_Option[i],8);
			for(j=0; j<8;j++)
			{
				eeprom_write_byte((uint8_t*)(EE_OPTS+(i*8)+j),g_Buffer[j]);
			}
		}
	}

	if(n == 2 || n == 4)
	{
		k=0;
		for(i=0; i<GetScrCount();i++)
		{
			for(j=0;j<4;j++)
			{
				char* Line = GetScr(i,j);
				for(l=0;Line[l] != 0;l++)
				{
					eeprom_write_byte((uint8_t*)(EE_END-k),Line[l]);
					k++;
				}
				eeprom_write_byte((uint8_t*)(EE_END-k),0);
				k++;
			}
			eeprom_write_byte((uint8_t*)(EE_END-k),GetScrAction(i));
			k++;
		}

		eeprom_write_byte((uint8_t*)(EE_END-k),0xFF);
	}
}
